<?

require_once "DB.php";
require_once "Cache.php";
require_once "HTTP/Request.php";
require_once "openssl.inc";
require_once "HTTP/Request/Listener.php";

define("BLZ_TOKEN_UNPROCESSED", (int) 0);
define("BLZ_TOKEN_PROCESSING", (int) 1);
define("BLZ_TOKEN_PROCESSED", (int) 2);

define("BLZ_REPLY", "REPLY");
define("BLZ_QUERY", "QUERY");

BLZConfig("logfile","cache/server.log");
BLZConfig("version","0.4");
BLZConfig("encryption",true); // Encrypt inter-node communication
BLZConfig("propagation",true); // Allow communication with peers
BLZConfig("traceoverhead",false); // Log traffic overhead
BLZConfig("site_root", ( preg_match('#/$#',dirname($_SERVER["SCRIPT_FILENAME"])) ? dirname($_SERVER["SCRIPT_FILENAME"]) : dirname($_SERVER["SCRIPT_FILENAME"])."/" ) );
BLZConfig("html_root", ( preg_match('#/$#',dirname($_SERVER["SCRIPT_NAME"])) ? dirname($_SERVER["SCRIPT_NAME"]) : dirname($_SERVER["SCRIPT_NAME"])."/" ) );
BLZConfig("local_search_mrr",1.0); // Minimal Replication Ratio (0 <= x <= 1) 

function BLZTraceOverhead($direction, $size, $time, $method) {
	if (BLZConfig("traceoverhead") != true)
		return;
		
	$string = "\"".date("Y/m/d H:i:s")."\", \"".BLZConfig("session_id")."\", \"$_SERVER[SERVER_ADDR]\", \"$_SERVER[REMOTE_ADDR]\", \"$direction\", \"$size\", \"$time\", \"$method\"";
	error_log("$string\n", 3, "cache/overhead.log");
}

if (! function_exists('sem_get')) {
	function sem_acquire($sem_identifier) {
		$fp = fopen("cache/lock_${sem_identifier}.sem", "w+");

		if (!flock($fp, LOCK_EX)) {
			debug_log("Couldn't aquire lock 'cache/lock_${sem_identifier}.sem'");
			fclose($fp);
			return false;			
		}
		
		fclose($fp);
		
		return true;
	}

	function sem_get($key, $max_acquire = null, $perm = null, $auto_release = null) {
		return $key;
	}

	function sem_release($sem_identifier) {
		$fp = fopen("cache/lock_${sem_identifier}.sem", "w+");

		if (!flock($fp, LOCK_UN)) {
			debug_log("Couldn't release lock 'cache/lock_${sem_identifier}.sem'");
			fclose($fp);
			return false;			
		}
		
		fclose($fp);
		
		return true;
	}
}

function debug($string,$noremoteoutput = true) {
	if ($_SERVER["HTTP_X_REMOTE_NODE_ID"] && $noremoteoutput) 
		debug_log($string);
	else {
		echo "<pre>".print_r($string,true)."</pre>";
		flush();
	}
}

function debug_log($string, $file = null) {
	if ($file == null)
		$file = BLZConfig("logfile");
		
	error_log(date("Y/m/d H:i:s")." [".BLZConfig("session_id")."]: $_SERVER[HTTP_HOST] $string\n", 3, $file);
}

function BLZEncrypt($data,$publickey) {	
	$ossl = new OpenSSL;
	$ossl->set_publickey($publickey);
	$ossl->encrypt(serialize($data));
	
	return array("crypt" => $ossl->get_crypt(), "ekey" => $ossl->get_ekey());
}

function BLZDecrypt($encypted_data,$privatekey,$pass) {
	$crypt = $encypted_data["crypt"];
	$ekey = $encypted_data["ekey"];
	$ossl = new OpenSSL;
	$ossl->set_privatekey($privatekey, false, $pass);
	$ossl->decrypt($crypt, $ekey);
	$plain = $ossl->get_plain();
	$plain = unserialize($ossl->get_plain());

	return $plain;
}

function BLZMicrotime() {
	$parts = split(" ",microtime());
	$mtf = (float) $parts[0] + (float) $parts[1]; 
	return $mtf; 
}


function BLZServeRequests() {
	if ($_SERVER["HTTP_X_SESSION_TOKEN"] && $_SERVER["HTTP_X_METHOD"]) {
		$server = new BLZServer();
		$starttime = BLZMicrotime(); 
		$response = $server->serveRequest();
		if (PEAR::isError($response)) {
			switch ($response->getCode()) {
				case BLZ_TOKEN_PROCESSING:
				case BLZ_TOKEN_PROCESSED:
					break;
				default:
					debug_log("ERROR -- '".$response->getMessage()." (".$response->getCode().")");
			}
		}
		else {
			ob_start("ob_gzhandler");
			echo $response;
			$time = BLZMicrotime() - $starttime;

			BLZTraceOverhead(BLZ_REPLY, ob_get_length(),$time,$_SERVER["HTTP_X_METHOD"]);
		}
		exit();
	}
}

function &BLZConfig($name,$value = null) {
	if ($value === null)
		return $GLOBALS["BLZ_CONFIG"][$name];
	else
		$GLOBALS["BLZ_CONFIG"][$name] = $value;
}

function BLZSetDB($dsn) {
	$db =& DB::connect($dsn);
	
	if (PEAR::isError($db))
		return new PEAR_Error($db->getMessage(), 0, PEAR_ERROR_RETURN);		

	BLZConfig("db",$db);
}

function BLZEnableCache($cache_dir = "cache") {
	$cache = new Cache('file', array('cache_dir' => $cache_dir));
	$cache->container->fileLocking = true;
	BLZConfig("cache",$cache);	
}

function &BLZGetFriends() {
	if ($friends = BLZConfig("friends"))
		return $friends;
		
	$db =& BLZConfig("db");
	$friends = $db->getAssoc("SELECT * FROM `blz_nodes` WHERE `private_key` = ''",false,null,DB_FETCHMODE_ASSOC);
	
	if (PEAR::isError($friends))
		return new PEAR_Error($friends->getMessage()."<br/>".$db->last_query,0,PEAR_ERROR_DIE);
		
	BLZConfig("friends",$friends);
	return $friends;
}

function BLZPrivateKey() {
	if ($private_key = BLZConfig("private_key"))
		return $private_key;
		
	$db =& BLZConfig("db");
	$private_key = $db->getOne("SELECT `private_key` FROM `blz_nodes` WHERE `private_key` != ''");
	
	if (PEAR::isError($private_key))
		return new PEAR_Error($private_key->getMessage()."<br/>".$db->last_query,0,PEAR_ERROR_DIE);
		
	BLZConfig("private_key",$private_key);
	return $private_key;
}

function BLZPrivatePass() {
	if ($private_pass = BLZConfig("private_pass"))
		return $private_pass;
		
	$db =& BLZConfig("db");
	$private_pass = $db->getOne("SELECT `private_pass` FROM `blz_nodes` WHERE `private_pass` != ''");
	
	if (PEAR::isError($private_pass))
		return new PEAR_Error($private_pass->getMessage()."<br/>".$db->last_query,0,PEAR_ERROR_DIE);
		
	BLZConfig("private_pass",$private_pass);
	return $private_pass;
}

function BLZNodeId() {
	if ($id = BLZConfig("id"))
		return $id;
		
	$db =& BLZConfig("db");
	$id = $db->getOne("SELECT `id` FROM `blz_nodes` WHERE `private_key` != ''");
	
	if (PEAR::isError($id))
		return new PEAR_Error($id->getMessage()."<br/>".$db->last_query,0,PEAR_ERROR_DIE);
		
	BLZConfig("id",$id);
	return $id;
}

// DEBUG function
function BLZNodeId2URI($id) {
	$db =& BLZConfig("db");
	$uri = $db->getOne("SELECT `server_uri` FROM `blz_nodes` WHERE `id` = ?",array($id));
	
	if (PEAR::isError($uri))
		return new PEAR_Error($uri->getMessage()."<br/>".$db->last_query,0,PEAR_ERROR_DIE);
		
	return $uri;
}

function &BLZLoadIdAttributes($id,$mdate = null,$usec = null) {
	$p = (bool) BLZConfig("propagation");
	BLZConfig("propagation",false); // Temporarily disable propagation
	$object =& BLZLoadId($id,$mdate,$usec);
	BLZConfig("propagation",$p); // Temporarily disable propagation
	return $object;
}

function &BLZLoadId($id,$mdate = null,$usec = null) {
	return BLZObject::loadId($id,$mdate,$usec);
}

function &BLZLoadURI($uri) {
	return BLZObject::loadURI($uri);
}

class BLZConcurrentRequestQueue extends PEAR {	
	var $_response_queue = array();
	var $_fps = array();
	var $_saveBody = array();
	var $_entries = 0;
	
	function process() {
		foreach ($this->_response_queue as $index => $response) {
			$err = $this->_response_queue[$index]->process($this->_saveBody[$index]);

			if (PEAR::isError($err)) {
				debug_log("Failed to connect to '".$response->_sock->addr."'");
				continue;
//            	return $err;
			}

            $response->_sock->disconnect();
        }
	}

	function _BLZConcurrentRequestQueue() {
		foreach ($this->_fps as $index => $fp) {
			if ($fp != null)
				$this->_response_queue[$index]->_sock->disconnect();
		}		
	}	

	function addResponse(&$response,$saveBody) {
		$this->_response_queue[] =& $response;
		$this->_fps[] =& $response->_sock->fp;
		$this->_saveBody[] = $saveBody;
		$index = $this->_entries;
		$this->_entries++;
		return $index; 
	}

    function getResponseCode($index = -1) {
        return isset($this->_response_queue[$index]->_code) ? $this->_response_queue[$index]->_code : false;
    }

    function getResponseHeader($index = -1, $headername = null) {
        if (!isset($headername)) {
            return isset($this->_response_queue[$index]->_headers)? $this->_response_queue[$index]->_headers: array();
        } else {
            $headername = strtolower($headername);
            return isset($this->_response_queue[$index]->_headers[$headername]) ? $this->_response_queue[$index]->_headers[$headername] : false;
        }
    }

    function getResponseBody($index = -1) {
    	if ($index > -1)
    		$list = array($this->_response_queue[$index]);
    	else
    		$list = $this->_response_queue;

		$result = array();
		foreach ($list as $response) {
			if (strlen($response->_body) > 0) {
				$body = $response->_body;
				$body = unserialize(urldecode($body));
		
				if (BLZConfig("encryption")) {
					$private_key = BLZPrivateKey();
					$private_pass =  BLZPrivatePass();
					$body = BLZDecrypt($body,$private_key,$private_pass);
				}
				
				$result[] = $body;
//				debug_log("response processed: ".$body[0]);
			}
		}
		
		return $result;
    }

    function getResponseCookies($index = -1) {
    	if ($index > -1)
        	return isset($this->_response_queue[$index]->_cookies) ? $this->_response_queue[$index]->_cookies : false;
        else {
        	for ($index = 0; $index < $this->_entries; $index++)
        		$result[] = isset($this->_response_queue[$index]->_cookies) ? $this->_response_queue[$index]->_cookies : false;
        	
        	return $result;
        }
    }
}

// NB: Does not allow redirects 
class BLZConcurrentRequest extends HTTP_Request {
	var $_response_queue = null;
	var $_queue_index = -1;

    function setResponseQueue(&$responsequeue) {
    	$this->_response_queue =& $responsequeue;
    }

    function _buildRequest() {
    	$request = HTTP_Request::_buildRequest();
    	$this->requestsize = strlen($request);
    	return $request;
    }

    function sendRequest($saveBody = true) {
        if (!is_a($this->_url, 'Net_URL')) {
            return PEAR::raiseError('No URL given.');
        }

        $host = isset($this->_proxy_host) ? $this->_proxy_host : $this->_url->host;
        $port = isset($this->_proxy_port) ? $this->_proxy_port : $this->_url->port;

        // 4.3.0 supports SSL connections using OpenSSL. The function test determines
        // we running on at least 4.3.0
        if (strcasecmp($this->_url->protocol, 'https') == 0 AND function_exists('file_get_contents') AND extension_loaded('openssl')) {
            if (isset($this->_proxy_host)) {
                return PEAR::raiseError('HTTPS proxies are not supported.');
            }
            $host = 'ssl://' . $host;
        }

        // magic quotes may fuck up file uploads and chunked response processing
        $magicQuotes = ini_get('magic_quotes_runtime');
        ini_set('magic_quotes_runtime', false);

        // If this is a second request, we may get away without
        // re-connecting if they're on the same server
        $err = $this->_sock->connect($host, $port, null, $this->_timeout, $this->_socketOptions);
        PEAR::isError($err) or $err = $this->_sock->write($this->_buildRequest());

        if (!PEAR::isError($err)) {
            if (!empty($this->_readTimeout)) {
                $this->_sock->setTimeout($this->_readTimeout[0], $this->_readTimeout[1]);
            }

            $this->_notify('sentRequest');

            // Add the response object to our listen list
        	$this->_response =& new HTTP_Response($this->_sock, $this->_listeners);
        	$this->_queue_index = $this->_response_queue->addResponse($this->_response,($this->_saveBody && $saveBody));
        }

        ini_set('magic_quotes_runtime', $magicQuotes);

        if (PEAR::isError($err)) {
            return $err;
        }

        return true;
    }
}

// TODO add safety features to serialization and unserialization
class BLZObject extends PEAR {
	var $getprops = true;

	/**
	 * @static
	 */ 
	function getTTL() {
		return 3600; // Replicas are good for 1 hour
	}
	
	function getURI() {
		$db = & BLZConfig("db");
		$uri = array ();
		$id = $this->id;
		$uri[] = rawurlencode($this->urn);
		
		while (true) {
			$id = $db->getOne("SELECT `parent_id` FROM `blz_attributes` WHERE `id` = ? AND `deleted` = ?", array ($id, 0));

			if ($id == "")
				break;
			else
				if ($id === null)
					return new PEAR_Error("Resource ($id) has an invalid parent", 0, PEAR_ERROR_RETURN);

			$name = $db->getOne("SELECT `urn` FROM `blz_attributes` WHERE `id` = ? AND `deleted` = ?", array ($id, 0));

			$uri[] = rawurlencode($name);
		}

		$uri = array_reverse($uri);
		$uri = "/".join("/", $uri);

		return $uri;
	}

	/**
	 * @static
	 */ 
	function &loadURI($uri) {
		$cache =& BLZConfig("cache");

		if ($cache) {
			$cache_id = $cache->generateID(md5($uri));
			if ($object_id =& $cache->get($cache_id)) {
				return BLZLoadId($object_id);
			}
		}

		$db =& BLZConfig("db");
		$path = split("/", $uri);
		$parent = "";
		$path_sofar = "";
		$remaining_path = "";

		foreach ($path as $urn) {
			if (!($urn = trim($urn)))
				continue;

			$path_sofar .= "/$urn";

/*
			if (get_magic_quotes_gpc() == 1) // XXX: Needed?
				$urn = stripslashes($urn);
*/

			$res = $db->query("SELECT `id` FROM `blz_attributes` WHERE `parent_id` = ? AND `urn` = ? AND `deleted` = ?", array ($parent, $urn, 0));
			$row =& $res->fetchRow(DB_FETCHMODE_ASSOC);
			$id = $row["id"];
			$res->free();

			if ($id == null)  // loadResourceUri failed 
				return null;

			$parent = $id;
		}

		$object =& BLZLoadId($id);
		
		if (!$object)
			return null;

		if ($cache) {
			$cache->save($cache_id, $id, $object->getTTL(),md5($uri));
		} 

		return $object; // loadID also updates the cache
	}


	/**
	 * @static
	 */ 
	function &loadId($id,$mdate,$usec) {
		$db =& BLZConfig("db");
		$query = "SELECT * FROM `blz_attributes` WHERE `id` = '$id' AND `deleted` = 0";
		$db->modifyLimitQuery($query, 0, 1);
		$res =& $db->query($query);

		if (PEAR::isError($res)) {
			return new PEAR_Error($res->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
		}		

		$row =& $res->fetchRow(DB_FETCHMODE_ASSOC);
		$res->free();
		
		if (!$row)
			return null;

		$class = $row["class"];
		unset($row["class"]);
		
		if (preg_match("/^_/",$class)) {
			$class = substr($class,1);

			if ($class == "")
				$class = "BLZObject";
				 
			$object = new $class();
			$object->getprops = false;
		}
		else		
			$object = new $class();
		
		foreach ($row as $key => $value)
			$object->$key = $value;
						
		// Check the freshness of the local object
		if (((time() - $object->udate) > $object->getTTL()) && $object->replica == 0) // Replica's never become "stale"
			$stale = true;
		else
			$stale = false;

		if ($object->getprops) {			
			if (BLZConfig("propagation") == true && $object->replica == 0) {
				$err = BLZClient::broadcastFetchReplica($object->id, $object->mdate, $object->usec);
		
				if (PEAR::isError($err))
					return new PEAR_Error($err->getMessage(), 0, PEAR_ERROR_RETURN);
					
				if (is_a($err,"BLZObject")) // A new version came from our friends
					$object =& $err;
			}			
			else if ($object->replica == 1) {
				$mdate = (int) $object->mdate;
				$usec = (int) $object->usec;
				$query = "SELECT * FROM `blz_class_$class` WHERE `id` = '$id' AND `mdate` = $mdate AND `usec` = $usec";
				$db->modifyLimitQuery($query, 0, 1);
				$res =& $db->query($query);
		
				if (PEAR::isError($res)) {
					return new PEAR_Error($res->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
				}		
		
				$row =& $res->fetchRow(DB_FETCHMODE_ASSOC);
				$res->free();
				foreach ($row as $key => $value)
					$object->$key = $value;
			}
		}
		else {
			if (BLZConfig("propagation") == true && $stale) {
				// TODO: Check around for a fresh version of the attributes
			} 	
		}
		
		$object->postLoad();

		return $object;
	}
	
	function postLoad() {
	}
	
	function preSave() {
		return true; 
	}
	
	function makeId() {
		return md5(microtime()." ".rand(1,100000));
	}
	
	function save($object = null) {
		$db =& BLZConfig("db");
		$cache =& BLZConfig("cache");

		if (is_object($object))
			$this =& $object;

//		debug("BLZObject::save($this->urn)");
			
		$class = get_class($this);
			
		$parent_id = ($this->parent_id ? $this->parent_id : "");

		// TODO verify & clean up urns
		
		if (PEAR::isError($presave = $this->preSave()))
			return $presave;
				
		if (!$object) {
			list($usec, $mdate) = explode(" ", microtime());
			$this->usec = (int) substr($usec,2,6);
			$this->mdate = (int) $mdate;
			$this->replica = (int) 1;
		}
		
		if (!$this->id) {
			$this->id = BLZObject::makeId();
			$new = true;
			
			// Check for overwrite
			$overwrite = $db->getOne("SELECT `id` FROM `blz_attributes` WHERE `urn` = ? AND `parent_id` = ? AND `deleted` = 0",
				array($this->urn,$parent_id));
			
			if ($overwrite)
				return new PEAR_Error("Save failed. Object would overwrite object ($overwrite)", 0, PEAR_ERROR_RETURN);
		}
		else if ($cache) {
			$cache_id = $cache->generateID($this->id);
			if ($cache->isCached($cache_id))
				$cache->flush($this->id); // id here is actually the group, not the cache_id 
		}
		
		
		$fields = array(
			"id" => $this->id,
			"parent_id" => $parent_id,
			"mdate" => $this->mdate,
			"usec" => $this->usec,
			"udate" => $this->mdate,
			"urn" => $this->urn,
			"class" => ($this->getprops ? $class : "_$class"),
			"replica" => (int) $this->replica,
			"deleted" => (int) $this->deleted,
			"system" => (int) $this->system,
		);
		
		// Mark other versions as deleted
		$res = $db->autoExecute("blz_attributes", array("deleted" => 1), DB_AUTOQUERY_UPDATE, "`id` = '$this->id'");

		if (PEAR::isError($res))
			return new PEAR_Error($res->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
			
		if ($object) {
			// Delete this version if it exists, allowing save to update objects
			$res = $db->query("DELETE FROM `blz_attributes` WHERE `id` = ? AND `mdate` = ? AND `usec` = ?", 
					array($this->id,$this->mdate,$this->usec));
	
			if (PEAR::isError($res))
				return new PEAR_Error($res->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);

			if ($this->getprops && $this->replica == 1) {
				$res = $db->query("DELETE FROM `blz_class_$class` WHERE `id` = ? AND `mdate` = ? AND `usec` = ?", 
						array($this->id,$this->mdate,$this->usec));

				if (PEAR::isError($res))
					return new PEAR_Error($res->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
			}
		}

		// Insert the new version
		$res = $db->autoExecute("blz_attributes", $fields);
		
		if (PEAR::isError($res))
			return new PEAR_Error($res->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);

		if ($this->getprops && $this->replica == 1) {
			$fields = array(
				"id" => $this->id,
				"mdate" => $this->mdate,
				"usec" => $this->usec,
			);
			$p = BLZClass::getProperties($class);
			foreach ($p as $id => $data) {
				list($propname,$type) = split(",",$data["urn"],2);
				$fields[$propname] = $this->$propname;
			}

			$res = $db->autoExecute("blz_class_$class", $fields);

			if (PEAR::isError($res))
				return new PEAR_Error($res->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);	
		}
						
		if (BLZConfig("propagation") == true) { // Only propagate changes if we're not already doing it. Prevent infinite propagation loops
			$err = BLZClient::broadcastUpdateAttributes($this->id, $parent_id, $this->mdate, $this->usec, $this->urn, $class, $this->system);

//			debug("BLZobject::save($this->urn) -- propagate attributes");

			if (PEAR::isError($err))
				return new PEAR_Error($err->getMessage(), 0, PEAR_ERROR_RETURN);	

			if ($this->getprops) {
//				debug("BLZobject::save($this->urn) -- propagate properties");
				$err = BLZClient::sendUpdateProperties($fields);
	
				if (PEAR::isError($err))
					return new PEAR_Error($err->getMessage(), 0, PEAR_ERROR_RETURN);	
			}
		}
		
		$this->postSave();

		return $this->id;
	}
	
	function postSave() {
	}
	
	function delete() {
		$db =& BLZConfig("db");
		$id = $this->id;
		$db->autoExecute("blz_attributes", array("deleted" => 1),DB_AUTOQUERY_UPDATE,"`id` = '$id'");

		if (BLZConfig("propagation") == true) {
			$err = BLZClient::broadcastDelete($this->id);

			if (PEAR::isError($err))
				return new PEAR_Error($err->getMessage(), 0, PEAR_ERROR_RETURN);
		}

		return true;
	}
}

class BLZObjectList extends PEAR {
	function property($propname,$value,$operator = "=") {
		$classes = array ();
		$propertytable = BLZClass::getPropertyTable();

		$db = & BLZConfig("db");
		$parts = split("\.", $propname);

		if (count($parts) == 2) {
			$classes[] = strtolower($parts[0]);
			$propname = $parts[1];
		} 
		else if (count($parts) == 1) {
			$proptype = null;

			foreach ($propertytable as $class => $proplist) {
				foreach ($proplist as $prop) {
					if ($prop["name"] == $propname) {
						$type = $prop["type"];

						if (($proptype != null) && ($proptype != $type))
							return new PEAR_Error("Ambiguous property name \"$propname\". "."Multiple properties of differing types have this name.", 0, PEAR_ERROR_RETURN);

						$proptype = $type;
						$classes[] = $class;
					}
				}
			}
		} 
		else
			return new PEAR_Error("Invalid property name \"$propname\"", 0, PEAR_ERROR_RETURN);

		if (count($this->_classes) > 0) {
			$this->_classes = array_unique(array_intersect($this->_classes, $classes));
			if (count($this->_classes) == 0)
				return new PEAR_Error("The conditions for this query are impossible to meet", 0, PEAR_ERROR_RETURN);			
		} else
			$this->_classes = $classes;

		foreach ($classes as $class) {
			$this->properties[$class][$propname][$operator] = $value;
		}

		return true;
	}

	function attribute($name,$value,$operator = "=") {
		$this->attributes[$name][$operator] = $value;
	}
	
	function getResult($group = false,$ttl = 60) {		
		// TODO: TTL should be lowest common denominator for TTL's in result set

		if ($_SERVER["HTTP_X_SESSION_TOKEN"])
			BLZConfig("cachedresult",false);
				
		$cache =& BLZConfig("cache");

		if ($cache) {
			$cache_id = $cache->generateID(array($this->attributes,$this->properties));
			if ($result =& $cache->get($cache_id)) {
				debug_log("Reused cached ObjectList ($cache_id)");			
				return $result;
			}
		}

		$db =& BLZConfig("db");

		$query = "";
		$where = array();

		if (!$this->attributes["deleted"])
			$this->attributes["deleted"]["="] = (int) 0;
			
		if (!$this->attributes["system"])
			$this->attributes["system"]["="] = (int) 0;

		if (count($this->properties) > 0)
			$this->attributes["replica"]["="] = (int) 1;

		foreach($this->attributes as $name => $comparison) {
			foreach($comparison as $operator => $value) {
				if (strtolower($operator) == "like")
					if (!preg_match("/%/",$value))
						$value = "%$value%";

				$where[] = "`blz_attributes`.`$name` $operator ".$db->quoteSmart($value);
				$debug_comparisons[] = "attribute.$name $operator $value";
			}	
		}	

		if (count($this->properties) > 0) {
			$mrr = BLZConfig("local_search_mrr");
			
			$result = array();
			$replication_level = array();
			foreach ($this->properties as $class => $prop) {
				$where2 = array();
				if (!in_array($class,$this->_classes))
					continue;
				foreach ($prop as $name => $comparison) {
					foreach ($comparison as $operator => $value) {
						if (strtolower($operator) == "like")
							if (!preg_match("/%/",$value))
								$value = "%$value%";

						$where2[] = "`blz_class_$class`.`$name` $operator ".$db->quoteSmart($value);

						$debug_comparisons[] = "$class.$name $operator $value";
					}
				}
				
				$query = "SELECT `blz_attributes`.`id`, `parent_id`, `blz_attributes`.`mdate`, `blz_attributes`.`usec`, `udate`, `urn`, `class`, `replica`, `deleted`, `system` " .
						"FROM `blz_attributes`, `blz_class_$class` WHERE ";
				$where2 = array_merge($where,$where2); 
				$query .= join(" AND ",$where2);
				$query .= " AND `blz_attributes`.`mdate` = `blz_class_$class`.`mdate`"; 
				$query .= " AND `blz_attributes`.`usec` = `blz_class_$class`.`usec`";				
				$query .= " ORDER BY `blz_attributes`.`mdate` ASC, `blz_attributes`.`usec` ASC"; // if group == false, getAssoc means that we get the newest version of each object			
				$result2 = $db->getAssoc($query, true, null, DB_FETCHMODE_ASSOC, $group);
				
				if (PEAR::isError($result2)) 
					return new PEAR_Error($result2->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
	
				$result = array_merge($result,$result2);
								
				$total_objects = (int) $db->getOne("SELECT COUNT(`id`) FROM `blz_attributes` WHERE `class` = '$class'");
				$total_replicas = (int) $db->getOne("SELECT COUNT(`id`) FROM `blz_attributes` WHERE `class` = '$class' AND `replica` = 1");			
				$replication_level["objects"] += $total_objects;
				$replication_level["replicas"] += $total_replicas;
			}
			
			if (BLZConfig("propagation") == true) {
				$local_replication = $replication_level["replicas"] / $replication_level["objects"];
				if ($local_replication < $mrr) {
					$remote_results = BLZClient::broadcastRangeQuery(array($this->attributes,$this->properties,$this->order,$this->limit));
					if (PEAR::isError($remote_results))
						return new PEAR_Error($remote_results->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
	
					$result = BLZObjectList::resultMerge($result,$remote_results);
					
					// TODO: detect if local attributes are outdated compared to receved resultset
				}
			}
		}
		else {
			$query = "SELECT * FROM `blz_attributes` WHERE ";
			$query .= join(" AND ",$where);	
			$query .= " ORDER BY mdate ASC, usec ASC"; // if group == false, getAssoc means that we get the newest version of each object			
			$result = $db->getAssoc($query, true, null, DB_FETCHMODE_ASSOC, $group);
			
			if (PEAR::isError($result)) 
				return new PEAR_Error($result->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
		}	

		// Only cache object lists that are not property queries 
		if (count($this->properties) == 0) { 
			debug_log("Caching ObjectList ($cache_id) [".join(",",$debug_comparisons)."]");
			$csr = $cache->save($cache_id, $result, $ttl);
		}
		
		return $result;		
	}
	
	function resultMerge($result_a,$result_b) {
		if (count($result_a) < count($result_b)) {
			$shortest =& $result_a;
			$longest =& $result_b;
		}
		else {
			$shortest =& $result_b;
			$longest =& $result_a;
		}

		$result = array();

		if(!is_array($shortest)) {
			return $longest;
		}
		
		foreach (array_keys($shortest) as $id) {
			if ($longest[$id]) {				
				if ($longest["mdate"] > $shortest["mdate"]) {
					// Do nothing. The version in $longest is better than the one in $shortest 
				}
				else if ($longest["mdate"] < $shortest["mdate"]) { 
					$longest[$id] = $shortest[$id];
				}	
				else if ($longest["usec"] >= $shortest["usec"]) { 
					// Do nothing. The version in $longest is better than the one in $shortest 
				}	
				else { // $longest["usec"] < $shortest["usec"]
					$longest[$id] = $shortest[$id];
				}	
			}
			else
				$longest[$id] = $shortest[$id];
		} 

		return $longest;
	}
}

class BLZPage extends BLZObject {
	var $default_properties = array("text,text");
	
	function BLZPage() {
		$this->smarty_template_dir = 'templates';
		$this->smarty_compile_dir = 'templates_c';
		$this->smarty_cache_dir = 'cache';
		$this->smarty_config_dir = 'configs';
		$this->smarty_caching = 2; // lifetime is per cache
		$this->smarty_cache_lifetime = $this->getTTL();
		$this->template = 'default.tpl';
		$this->computed_values = array();
	}
	
	function render() {
		$smarty = new Smarty();
		$smarty->template_dir = $this->smarty_template_dir;
		$smarty->compile_dir = $this->smarty_compile_dir;
		$smarty->cache_dir = $this->smarty_cache_dir;
		$smarty->config_dir = $this->smarty_config_dir;
		$smarty->caching = $this->smarty_caching; // lifetime is per cache
		$smarty->cache_lifetime = $this->smarty_cache_lifetime;
		
		if(!$smarty->is_cached($this->template,$this->id)) {	
			foreach ($this as $key => $value) {
				if ($key == "smarty")
					continue;					
				$smarty->assign($key, $value);
			}

			foreach ($this->computed_values as $function)
				$smarty->assign($function, $this->$function());
		}
		
		$smarty->display($this->template,$this->id);
	}
}

class BLZClass extends BLZObject {
	var $getprops = false;
	var $system = 1;
	
	function BLZClass($name = null, $props = null) {
		$this->PEAR();
		if ($name)
			$this->urn = $name;
			
		if (is_array($props)) {
			$this->_makeprops = $props; 
		}
		
		// TODO: respect $foo->default_properties
	}
	
	function getProperties($class) {
		$list = new BLZObjectList(); 
		$list->attribute("class","_BLZClass");
		$list->attribute("system",1);
		$list->attribute("urn",$class);
		$id = key($list->getResult());
		$list = new BLZObjectList(); 
		$list->attribute("class","_BLZProperty");
		$list->attribute("parent_id",$id);
		$list->attribute("system",1);
		return $list->getResult();
	}
	
	function getPropertyTable($ttl = 5) {
		$cache =& BLZConfig("cache");

		if ($cache) {
			$cache_id = $cache->generateID("getPropertyTable");
			if ($propertytable = $cache->get($cache_id))
				return $propertytable;
		}

		$propertytable = array();		
		$list = new BLZObjectList(); 
		$list->attribute("class","_BLZClass");
		$list->attribute("system",1);
		$classlist = $list->getResult(); 

		foreach($classlist as $id => $data) {
			$class = $data["urn"];
			$list = new BLZObjectList(); 
			$list->attribute("class","_BLZProperty");
			$list->attribute("parent_id",$id);
			$list->attribute("system",1);
			$proplist = $list->getResult();
			
			foreach($proplist as $pid => $pdata) {
				list($pname, $ptype) = explode(",", $pdata["urn"],2);
				$property = array("name" => $pname,"type" => $ptype);
				$propertytable[$class][] = $property;
			}			
		}

		if ($cache)
			$cache->save($cache_id, $propertytable, $ttl); 

		return $propertytable;
	}
	
	function preSave() {
		$this->urn = strtolower($this->urn);
		
		if ($this->_original_urn && $this->_original_urn != $this->urn)
			return new PEAR_Error("Save failed. Classes may not be renamed.", 0, PEAR_ERROR_RETURN);	
		
		// TODO fail if class is nonexistant.		
		
		return true; 
	}
	
	function save($object = null) {
		$db =& BLZConfig("db");

		if (is_object($object)) {
			$this =& $object;
			$exists = BLZLoadIdAttributes($object->id);
			$new = (bool) (!$exists);
			unset($exists);
		}
		else {
			$new = (bool) (!$this->id);
		}
			
//		debug("BLZClass::save($this->urn)");

		if (PEAR::isError($id = parent::save()))
			return $id;

		if ($new) {
			$query = "CREATE TABLE IF NOT EXISTS `blz_class_".$this->urn."` (" .
				"`id` varchar( 32 ) NOT NULL default ''," .
				"`mdate` int( 10 ) NOT NULL default '0'," .
				"`usec` int( 6 ) NOT NULL default '0'," .
				"PRIMARY KEY ( `id` , `mdate` , `usec` )" .
				") ENGINE = MYISAM";

			$result = $db->query($query);
			if (PEAR::isError($result))
				return new PEAR_Error("Save failed. There can only be one class with the name '$this->urn'.", 0, PEAR_ERROR_RETURN);
		}

		if (is_array($this->_makeprops)) {
			foreach ($this->_makeprops as $urn) {
//				debug("BLZClass::save($this->urn) - make property '$urn'");

				$property = new BLZProperty();
				$property->urn = $urn;
				$property->parent_id = $this->id;
				$prop_id = $property->save();

/*
				if (PEAR::isError($prop_id)) {
//					debug("BLZClass::save($this->urn) -- save failed: ".$prop_id->getMessage());
					return new PEAR_Error("Save failed: ".$prop_id->getMessage(), 0, PEAR_ERROR_RETURN);
				}
*/
			}
			unset($this->_makeprops);			
		}

		return $id;
	}
	
	function postLoad() {
		$this->urn = strtolower($this->urn);
		$this->_original_urn = $this->urn; 
	}
	
	function delete() {
		// TODO - flush cache to ensure all objects are deleted

		// Delete all objects of this type
		$list = new BLZObjectList();
		$list->attribute("class",$this->urn);
		$objects = $list->getResult();
		foreach ($objects as $id => $data) {
			$tmp = BLZLoadIdAttributes($id);
			$tmp->delete();
		}

		// Delete all properties of this type
		$list = new BLZObjectList();
		$list->attribute("parent_id",$this->id);
		$list->attribute("class","_BLZProperty");
		$list->attribute("system",1);
		$objects = $list->getResult();
		foreach ($objects as $id => $data) {
			$tmp = BLZLoadIdAttributes($id);
			$tmp->delete(false);
		}

		$class = $this->urn;
		parent::delete();

		$db =& BLZConfig("db");
		$result = $db->query("DROP TABLE `blz_class_".$class."`");

		if (PEAR::isError($result))
			return new PEAR_Error($result->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
	}
}

class BLZProperty extends BLZObject {
	var $getprops = false;
	var $system = 1;
	var $name;
	var $type;

	function BLZProperty($propertyname = null,$propertytype = null,$class_id = null) {
		if ($propertyname && $propertytype && $class_id) {
			$this->urn = "$propertyname,$propertytype";
			$this->parent_id = $class_id;
		}
	}
	
	function postLoad() {
		list($this->name, $this->type) = explode(",", $this->urn,2);
		$this->urn = strtolower($this->urn);
		$this->_original_urn = $this->urn; 
	}
	
	function preSave() {
		$this->urn = strtolower($this->urn);
		list($this->name, $this->type) = explode(",", $this->urn,2);
		
		if ($this->_original_urn && $this->_original_urn != $this->urn)
			return new PEAR_Error("Save failed. Properties may not be renamed.", 0, PEAR_ERROR_RETURN);	
		
		return true; 
	}
	
	function save($object = null) {
		$db =& BLZConfig("db");
			
		if (is_object($object)) {
			$this =& $object;
			$exists = BLZLoadIdAttributes($object->id);
			$new = (bool) (!$exists);
			unset($exists);
		}
		else {
			$new = (bool) (!$this->id);
		}

//		debug("BLZProperty::save($this->urn)");

		$err = parent::save();
		
		if (PEAR::isError($err))
			return new PEAR_Error($err->getMessage(), 0, PEAR_ERROR_RETURN);	

		if ($this->_original_urn && $this->_original_urn != $this->urn)
			return new PEAR_Error("Save failed. Properties may not be renamed.", 0, PEAR_ERROR_RETURN);

		if ($new) {
			list($this->name, $this->type) = explode(",", $this->urn,2);

			$parent_urn = $db->getOne("SELECT `urn` FROM `blz_attributes` WHERE `id` = ? AND `deleted` = 0",
				array($this->parent_id));
			if (PEAR::isError($parent_urn))
				return new PEAR_Error($parent_urn->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
			$result = $db->query("ALTER TABLE `blz_class_$parent_urn` ADD `$this->name` $this->type NOT NULL");
			if (PEAR::isError($result))
				return new PEAR_Error($result->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
		}
		
		return $err;
	}
	
	function delete($altertable = true) {
		if ($altertable) {
			$db =& BLZConfig("db");

			$parent_urn = $db->getOne("SELECT `urn` FROM `blz_attributes` WHERE `id` = ? AND `deleted` = 0",
				array($this->parent_id));
			if (PEAR::isError($parent_urn))
				return new PEAR_Error($parent_urn->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);

			$result = $db->query("ALTER TABLE `blz_class_$parent_urn` DROP `$this->name`");
			if (PEAR::isError($result))
				return new PEAR_Error($result->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
		}
		
		return parent::delete();
	}
}

class BLZToken extends PEAR {
	var $_aquired = false;
	
	function BLZToken($token = null, $auto_aquire = true) {
		parent::PEAR();

		if ($token == null)
			$token = md5(microtime());
		
		$this->token = $token;

		if ($auto_aquire == true) {				
			$result = $this->aquire();
			if (PEAR::isError($result))
				$this = new PEAR_Error($result->getMessage(), $result->getCode(), PEAR_ERROR_RETURN);				
		}
	}
	
	function _BLZToken() {
		if ($this->_aquired == true) {
			$this->release(); // Just a precaution
		}
	}
	
	function _aquireSempahore() {
		if (!($sem_id = & BLZConfig("sem_id"))) {
			$shm_key = ftok(__FILE__, 't'); // XXX: Fucks up if other semaphores are used in this file!
			$sem_id = sem_get($shm_key, 1, 0666, true); // Auto released semaphore
			BLZConfig("sem_id", $sem_id);
		}
	
		sem_acquire($sem_id);
	}
	
	function _releaseSempahore() {
		if (!($sem_id = & BLZConfig("sem_id")))
			new PEAR_Error('Invalid token semaphore', null, PEAR_ERROR_RETURN);
	
		sem_release($sem_id);
	}
	
	// Access should happen in sempaphor controlled blocks
	function _setStatus($status) {
		$db =& BLZConfig("db");
		$res = $db->query("DELETE FROM `blz_tokens` WHERE `token` = ?",array($this->token));

		if (PEAR::isError($res))
			return new PEAR_Error($res->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);

		$fields = array(
			"token" => $this->token,
			"status" => (int) $status,
			"origin" => "",
		);

		$res = $db->autoExecute("blz_tokens", $fields);
	
		if (PEAR::isError($res))
			return new PEAR_Error($res->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
		else
			return $res;
	}
	
	// Access should happen in sempaphor controlled blocks
	function _getStatus() {
		$db =& BLZConfig("db");
		$status = $db->getOne("SELECT `status` FROM `blz_tokens` WHERE `token` = ?",array($this->token));

		if (PEAR::isError($status))
			return new PEAR_Error($status->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);
		else
			return (int) $status;
	}

	function isAquired() {
		return $this->_aquired;
	}
	
	function aquire() {
		$this->_aquireSempahore();
		
		$status = $this->_getStatus();
		if (PEAR::isError($status))
			return new PEAR_Error($status->getMessage(), 0, PEAR_ERROR_DIE);

		if ($status == BLZ_TOKEN_UNPROCESSED) { 
			$status = BLZ_TOKEN_PROCESSING;
			$result = $this->_setStatus($status);

			if (PEAR::isError($result))
				return new PEAR_Error($result->getMessage(), 0, PEAR_ERROR_DIE);

			$this->_aquired = true;
			$this->_releaseSempahore();		
			return true;
		}

		$this->_aquired = false;
		$this->_releaseSempahore();		

		if ($status == BLZ_TOKEN_PROCESSING)
			return new PEAR_Error("The token was not aquired. It is already being processed.", BLZ_TOKEN_PROCESSING, PEAR_ERROR_RETURN);				
		else
			return new PEAR_Error("The token was not aquired. It has already been processed.", BLZ_TOKEN_PROCESSED, PEAR_ERROR_RETURN);
	}

	function release() {
		if ($this->_aquired === false)
			return new PEAR_Error("Unable to release token '$this->token'. It was never aquired.", 0, PEAR_ERROR_RETURN);
			
		$this->_aquireSempahore();
	
		$status = $this->_getStatus();

		if ($status == BLZ_TOKEN_PROCESSING) {
			$this->_setStatus(BLZ_TOKEN_PROCESSED);
			$return = true;
		}
		else
			$return = false;
	
		$this->_releaseSempahore();
		$this->_aquired = false;

		if ($return)
			return true;
		else	
			return new PEAR_Error("Unable to release token '$this->token'. It was not marked as being processed.", 0, PEAR_ERROR_RETURN);
	}
	
	function toString() {
		return $this->token;
	}
}

class BLZServerBase extends PEAR {
	function serveRequest() {
		// TODO verify remote Blitz version is older or equal to own version
		
		$token_string = $_SERVER["HTTP_X_SESSION_TOKEN"];
		$method = $_SERVER["HTTP_X_METHOD"];

//		debug_log("serveRequest($method,$token_string)");

		if (!method_exists($this,$method)) {
			header("X-Blitz-Error: Invalid method");
			return new PEAR_Error("Invalid method '$method'", 0, PEAR_ERROR_RETURN);
		}
		
		// Make a token object and auto aquire it
		$token = new BLZToken($token_string);

		if (PEAR::isError($token))
			return new PEAR_Error($token->getMessage(), $token->getCode(), PEAR_ERROR_RETURN);
		
		$payload = $_POST["payload"];
		$payload = unserialize(urldecode($payload));

		if (BLZConfig("encryption")) {
			$private_key = BLZPrivateKey();
			$private_pass =  BLZPrivatePass();
			$payload = BLZDecrypt($payload,$private_key,$private_pass);
		}

		if ($_SERVER["HTTP_X_BROADCAST"] == "true") {
			$waitforresponse = ($_SERVER["HTTP_X_WAITFORRESPONSE"] == "true" ? true : false);		

			$result = BLZClientBase::broadcast($method, $payload, $token, $waitforresponse);
		}		
		
		BLZConfig("propagation",false);		
		$result[] = $this->$method($payload,$token); // Call the requested method
		BLZConfig("propagation",true);

		$friends = BLZGetFriends();

		$response = $result;

		if (BLZConfig("encryption")) {		
			$target = $_SERVER["HTTP_X_REMOTE_NODE_ID"];
			$response = BLZEncrypt($response,$friends[$target]["public_key"]);
		}	

		$response = urlencode(serialize($response));

		// Release the token
		$token->release();

		return $response;
	}
}

class BLZServer extends BLZServerBase {
	function hello($payload,$token = null) {
		$friends =& BLZGetFriends();
		
		return "($_SERVER[HTTP_HOST]) RECEIVED (".print_r($payload,true).")";
	}
	
	function update_attributes($payload,$token = null) {
		$db =& BLZConfig("db");
		$class = $payload[5];	
		$object = new $class();
			
		// TODO only update if the recived version is newer

		$object->id = $payload[0];
		$object->parent_id = $payload[1];
		$object->mdate = (int) $payload[2];
		$object->usec = (int) $payload[3];
		$object->udate = time();
		$object->urn = $payload[4];
		$object->replica = (int) 0;
		$object->deleted = (int) 0;
		$object->system = (int) $payload[6];

		$err = call_user_func(array($class, 'save'),$object);
		
		if (PEAR::isError($err))
			return new PEAR_Error($err->getMessage(), 0, PEAR_ERROR_RETURN);
		else
			return "OK";
	}
	
	function update_properties($payload,$token = null) {
		$db =& BLZConfig("db");
		$object = BLZLoadId($payload["id"]);

		$fields = array(
			"id" => $payload["id"],
			"mdate" => (int) $payload["mdate"],
			"usec" => (int) $payload["usec"],
		);
		
		// Only proceed if the received props are fresher than the local one
		if ($object->mdate > $fields["mdate"] || ($object->mdate == $fields["mdate"] && $object->usec > $fields["usec"]))
			return "NEWER LOCAL VERSION";
					
		$class = get_class($object);

		foreach (BLZClass::getProperties($class) as $id => $data) {
			list($propname,$type) = split(",",$data["urn"],2);
			$fields[$propname] = $payload[$propname];
		}

		$res = $db->autoExecute("blz_class_$class", $fields);

		if (PEAR::isError($res))
			return new PEAR_Error($res->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);	

		$res = $db->autoExecute("blz_attributes", array("replica" => 1), DB_AUTOQUERY_UPDATE, 
			"`id` = '$payload[id]' AND `mdate` = '$payload[mdate]' AND `usec` = '$payload[usec]'");
		
		if (PEAR::isError($res))
			return new PEAR_Error($res->getMessage()."<br/>".$db->last_query, 0, PEAR_ERROR_RETURN);	

		return "OK";
	}

	function delete($payload,$token = null) {
		$object = BLZLoadId($payload);
		$err = $object->delete();

		if (PEAR::isError($err))
			return new PEAR_Error($err->getMessage(), 0, PEAR_ERROR_RETURN);
		else	
			return "OK";
	}
	
	function fetch_replica($payload,$token) {
		// TODO: Returning the object instance is wastefull. only return attribs and props in an array.
		
		$id = $payload[0];
		$mdate = (int) $payload[1]; 
		$usec = (int) $payload[2];
		$object = BLZLoadId($id);
		$object->origin = BLZNodeId();
		
		if (is_a($object,"BLZObject")) {
			if ($object->mdate == $mdate && $object->usec == $usec && $object->replica == 1) {				
				return $object; // Bingo!
			}
			else if (($object->mdate > $mdate ||
				($object->mdate == $mdate && $object->usec > $usec)) && 
				$object->replica == 1) 
			{
				return $object; // The local replica was actually newer than the requested?
			}
		}  

		// No luck here. Ask out friends
		$object = BLZClient::broadcastFetchReplica($id, $mdate, $usec, $token);
		
		if (PEAR::isError($object))
			return new PEAR_Error($object->getMessage(), 0, PEAR_ERROR_RETURN);

		return $object;
	}
	
	function range_query($payload,$token) {
		$list = new BLZObjectList();
		$list->attributes = $payload[0];
		$list->properties = $payload[1];
		$list->order = $payload[2];
		$list->limit = $payload[3];
		$list->_classes = array_keys($list->properties);
		$result = $list->getResult();
		
		if (BLZConfig("cachedresult") == true)
			return $result; // Short circuit
			
		$remote_result = BLZClient::broadcastRangeQuery($payload, $token);

		if (PEAR::isError($remote_result))
			return new PEAR_Error($remote_result->getMessage(), 0, PEAR_ERROR_RETURN);
			
		$result = BLZObjectList::resultMerge($result,$remote_result);

		return $result;
	}
}

class BLZClientBase extends PEAR {
	function _call_blocking($method, $payload, $token, $waitforresponse, $target, $broadcast) {
		$params = array();
		$friends =& BLZGetFriends();
	
		if (!$token->isAquired())
			return new PEAR_Error("Bad token '".print_r($token,true)."'", 0, PEAR_ERROR_RETURN);
		
		// TODO: Handle  $waitforresponse == false

		$server_uri = $friends[$target]["server_uri"];

		if (BLZConfig("encryption"))		
			$payload = BLZEncrypt($payload,$friends[$target]["public_key"]);	

		$payload = urlencode(serialize($payload));
		
		$request = new HTTP_Request($server_uri, $params);
		$request->setMethod(HTTP_REQUEST_METHOD_POST);
		$request->addHeader('User-Agent', 'Blitz/'.BLZConfig("version"));
		$request->addHeader('X-Session-Token', $token->toString());
		$request->addHeader('X-Method', $method);
		$request->addHeader('X-Broadcast', ($broadcast ? "true" : "false"));
		$request->addHeader('X-Remote-Node-Id', BLZNodeId());
		$request->addPostData("payload", $payload);

		$request->sendRequest();
	
		// TODO: Check for errors from remote server

		$response = $request->getResponseBody();
		$response = unserialize(urldecode($response));

		if (BLZConfig("encryption")) {
			$private_key = BLZPrivateKey();
			$private_pass =  BLZPrivatePass();
			$response = BLZDecrypt($response,$private_key,$private_pass);
		}

		return $response;
	}
	
	function _call(&$responsequeue, $method, $payload, $token, $waitforresponse, $target, $broadcast) {
		$params = array();
		$friends =& BLZGetFriends();
		$server_uri = $friends[$target]["server_uri"];
	
//		debug_log("_call($method,$server_uri,".$token->toString().")");

		if (!$token->isAquired())
			return new PEAR_Error("Bad token '".print_r($token,true)."'", 0, PEAR_ERROR_RETURN);
		
		// TODO: Handle  $waitforresponse == false


		if (BLZConfig("encryption"))		
			$payload = BLZEncrypt($payload,$friends[$target]["public_key"]);	

		$payload = urlencode(serialize($payload));

//		debug_log("_call($method,$target,".print_r($payload,true).")");
		
		$request = new BLZConcurrentRequest($server_uri, $params);
		$request->setResponseQueue($responsequeue);
		$request->setMethod(HTTP_REQUEST_METHOD_POST);
		$request->addHeader('User-Agent', 'Blitz/'.BLZConfig("version"));
		$request->addHeader('X-Session-Token', $token->toString());
		$request->addHeader('X-Method', $method);
		$request->addHeader('X-Broadcast', ($broadcast ? "true" : "false"));
		$request->addHeader('X-Remote-Node-Id', BLZNodeId()); // Debug?
		$request->addPostData("payload", $payload);

		$starttime = BLZMicrotime();	
		$request->sendRequest();
		$time = BLZMicrotime() - $starttime;	

		BLZTraceOverhead(BLZ_QUERY, $request->requestsize, $time, $method);

//		debug_log("sent request to '$server_uri'");

		// TODO: Check for errors from remote server
	}
	
	function call($method, $payload, $target, $token = null, $waitforresponse = true) {
		// TODO: gather rtt data and update friendslist
		if ($token == null) {		
			$token = new BLZToken();
			$release = true;
		}
		$responsequeue = new BLZConcurrentRequestQueue();
		$result = BLZClientBase::_call($responsequeue, $method, $payload, $token, $waitforresponse, $target, false);

		if ($waitforresponse) {	
			if (PEAR::isError($err = $responsequeue->process())) {				
				return new PEAR_Error("ERROR: ".$err->getMessage()." (".$err->getCode().")",$err->getCode(),PEAR_ERROR_DIE);
			}
		
			$result = $responsequeue->getResponseBody();
			
			if ($release)
				$token->release();
	
			return $result[0];
		}
		else
			return true;
	}
	
	function broadcast($method, $payload, $token = null, $waitforresponse = true) {
		// TODO: gather rtt data and update friendslist
		if ($token == null) {
			$token = new BLZToken();
			$releasetoken = true;
		}

		$responsequeue = new BLZConcurrentRequestQueue();
		$friends = array_keys(BLZGetFriends());		
		$result = array();
		
		foreach ($friends as $target) {
			if ($target == $_SERVER["HTTP_X_REMOTE_NODE_ID"])
				continue;
				
			$resp = BLZClientBase::_call($responsequeue, $method, $payload, $token, $waitforresponse, $target, true);
		}
		
		if ($waitforresponse) {
			if (PEAR::isError($err = $responsequeue->process()))
				return new PEAR_Error("ERROR: ".$err->getMessage()." (".$err->getCode().")",$err->getCode(),PEAR_ERROR_DIE);
			
			$result = array();
			foreach ($responsequeue->getResponseBody() as $response) {
				$result = array_merge($result,$response);
			}
			
			if ($releasetoken)
				$token->release();
				
			return $result;
		}
		else
			return true;
	}
}

class BLZClient extends BLZClientBase {
	function hello($target,$message = "Hello") {
		return parent::call("hello", $message, $target); 
	}

	function bc_hello($message = null) {
		return parent::broadcast("hello", $message);
	}
	
	function broadcastUpdateAttributes($id, $parent_id, $mdate, $usec, $urn, $class, $system) {
		return parent::broadcast("update_attributes", array($id, $parent_id, $mdate, $usec, $urn, $class, $system));
	}

	function sendUpdateProperties($fields) {
		$friends = array_keys(BLZGetFriends());
		$result = array();
		foreach($friends as $target) {
			$result[] = parent::call("update_properties", $fields, $target);
		}
		
		return $result;
	}
	
	function broadcastDelete($id) {
		return parent::broadcast("delete", $id, null, false);
	}
	
	function &broadcastFetchReplica($id, $mdate, $usec, $token = null) {
		/* 
		 * We don't use a regular broadcast here because we want the nodes along
		 * the path to update their own data based on the data that passes
		 * back through them upstream.
		 */
		if ($token == null) {
			$token = new BLZToken();
			$release = true;
		}

		$friends = array_keys(BLZGetFriends());
		$result = array();
		$responsequeue = new BLZConcurrentRequestQueue();
		
		foreach($friends as $target) {
			if ($target == $_SERVER["HTTP_X_REMOTE_NODE_ID"])
				continue;

			BLZClientBase::_call($responsequeue, "fetch_replica", array($id, $mdate, $usec), $token, true, $target, false);
		}

		if (PEAR::isError($err = $responsequeue->process()))
			return new PEAR_Error("ERROR: ".$err->getMessage()." (".$err->getCode().")",$err->getCode(),PEAR_ERROR_DIE);

		$tmp = $responsequeue->getResponseBody();
		foreach ($tmp as $objects) {
			$result = array_merge($result, $objects);
		}

		$bestmdate = 0;
		$bestusec = 0;
		$bestobject = null;
		foreach ($result as $object) {
			// filter response and update local version					
			if (is_a($object,"BLZObject")) {
				if ($object->mdate > $bestmdate || ($object->mdate == $bestmdate && $object->usec > $bestusec)) {
					$bestmdate = $object->mdate;
					$bestusec = $object->usec;
					$bestobject = $object;
				}
			}
		}

		// Save best replica locally
		if ($bestobject != null) {
			BLZConfig("propagation",false);
			$id = $bestobject->save($bestobject);
			BLZConfig("propagation",true);
			if (PEAR::isError($id))
				debug_log("save failed (".__FILE__.":".__LINE__.": ".$id->getMessage());
		}
		
		if ($release == true)
			$token->release();
											
		return $bestobject;
	}
	
	function broadcastRangeQuery($query, $token = null) {
		list($attributes,$properties,$order,$limit) = each($query);

		if ($token == null) {
			$token = new BLZToken();
			$release = true;
		}

		$friends = array_keys(BLZGetFriends());
		$result = array();
		$responsequeue = new BLZConcurrentRequestQueue();
		
		foreach($friends as $target) {
			if ($target == $_SERVER["HTTP_X_REMOTE_NODE_ID"])
				continue;

			BLZClientBase::_call($responsequeue, "range_query", $query, $token, true, $target, false);
		}

		if (PEAR::isError($err = $responsequeue->process()))
			return new PEAR_Error("ERROR: ".$err->getMessage()." (".$err->getCode().")",$err->getCode(),PEAR_ERROR_DIE);

		$tmp = $responsequeue->getResponseBody();

		foreach ($tmp as $objects) {
			$result = BLZObjectList::resultMerge($result, $objects[0]);
		}
		
		if ($release == true)
			$token->release();
											
		return $result;
	}
}


class BLZFile extends BLZObject {
	var $default_properties = array("file_uri,text","cache_urn,text");	
	var $file_uri; // A valid URI to the object, ie: "http://www.blitzcms.org/cache/e5j5aa5afefe778.jpg"
	var $filecache = "cache"; // The cache path, relative to the site_root	
	var $cache_urn = "";
	
	function BLZFile($uri = null) {
		$this->filecache = (preg_match('#/$#',$this->filecache) ? 
			$this->filecache : $this->filecache."/" );

		if ($uri)
			$this->upload($uri);
	}
		
	// Upload saves the file!!
	function upload($uri) {		
		if (!$this->urn)
			$this->urn = basename($uri);

debug_log("upload($uri)");
		
		$suffix = "";
		preg_match("#\.[^.]+$#",basename($uri),$suffix);
		$this->file_suffix = $suffix[0]; // TODO
		$this->replica = 1;

		if (!$this->cache_urn)	
			$this->cache_urn = BLZObject::makeId().$this->file_suffix;

debug_log("\$this->cache_urn = $this->cache_urn");
			
		if (!copy($uri,$this->getCachePath())) {
debug_log("copy() failed");
			return new PEAR_Error("Could not copy '$uri' to '".$this->getCachePath()."'.", 
				0, PEAR_ERROR_RETURN);
		}
		else {
			// TODO: hardcoding "http://" does allow https connections... not good
			$this->file_uri = "http://".$_SERVER["HTTP_HOST"].$this->getCacheURI();

debug_log("copy('$uri','".$this->getCachePath()."')");

			if (PEAR::isError($error = $this->save())) {
debug_log("save() failed");
				unlink($this->getCachePath());
				return $error;
			}			
		}
debug_log("upload complete");
	}
	
	function delete() {
		unlink($this->getCachePath());
		parent::delete();
	}
	
	function getCacheURI() {
		return BLZConfig("html_root").$this->filecache.$this->cache_urn;
	}
		
	function getCachePath() {
		return BLZConfig("site_root").$this->filecache.$this->cache_urn;
	}
	
	function isCached() {
		if (file_exists($this->getCachePath()))
			return true;
		else
			return false;
	}
	
	function render() {
		if ($this->isCached()) {
			header("Location: ".$this->getCacheURI());
			exit();
		}
		else {
			/*
			 * TODO: Temporary solution! The file should be streamed to the
			 * visitor simultaneously!
			 */ 
			$p = BLZConfig("propagation");
			BLZConfig("propagation",false);

			if (!PEAR::isError($error = $this->upload($this->file_uri))) {
				header("Location: ".$this->getCacheURI());
				exit();
			}

			BLZConfig("propagation",$p);
			return new PEAR_Error("File upload failed: ".$error->getMessage(), 0,
				PEAR_ERROR_RETURN);
		}
	}
}

?>
