<?

/*

CLASS OpenSSL

A wrapper class for a simple subset of the PHP OpenSSL functions. Use for public key encryption jobs.

=== Includes source code from many contributors to the PHP.net manual ===

....usage examples below....

Alex Poole 2005

php ~at~ wwwcrm.com

*/

DEFINE("OPEN_SSL_CONF_PATH", "/etc/ssl/openssl.cnf"); //point to your config file
DEFINE("OPEN_SSL_CERT_DAYS_VALID", 365); //1 year
DEFINE("OPEN_SSL_IS_FILE", 1);

class OpenSSL {

	var $privatekey; //resource or string private key
	var $publickey; //ditto public
	var $plaintext;
	var $crypttext;
	var $ekey; //ekey - set by encryption, required by decryption
	var $privkeypass; //password for private key
	var $csr; //certificate signing request string generated with keys
	var $config;

	function OpenSSL() {
		$this->config = array ("config" => OPEN_SSL_CONF_PATH);
	}

	function readf($path) {
		//return file contents
		$fp = fopen($path, "r");
		$ret = fread($fp, 8192);
		fclose($fp);
		return $ret;
	}

	//privatekey can be text or file path
	function set_privatekey($privatekey, $isFile = 0, $key_password = "") {
		if ($key_password)
			$this->privkeypass = $key_password;

		if ($isFile)
			$privatekey = $this->readf($privatekey);

		$this->privatekey = openssl_get_privatekey($privatekey, $this->privkeypass);
	}

	//publickey can be text or file path
	function set_publickey($publickey, $isFile = 0) {
		if ($isFile)
			$publickey = $this->readf($publickey);

		$this->publickey = openssl_get_publickey($publickey);
	}

	function set_ekey($ekey) {
		$this->ekey = $ekey;
	}

	function set_privkeypass($pass) {
		$this->privkeypass = $pass;
	}

	function set_plain($txt) {
		$this->plaintext = $txt;
	}

	function set_crypttext($txt) {
		$this->crypttext = $txt;
	}

	function encrypt($plain = "") {		
		if ($plain)
			$this->plaintext = $plain;

		$ekey = ""; // Make Eclipse STFU			
		openssl_seal($this->plaintext, $this->crypttext, $ekey, array ($this->publickey));

		$this->ekey = $ekey[0];
	}

	function decrypt($crypt = "", $ekey = "") {
		if ($crypt)
			$this->crypttext = $crypt;
		if ($ekey)
			$this->ekey = $ekey;

		openssl_open($this->crypttext, $this->plaintext, $this->ekey, $this->privatekey);
	}

	function do_csr($countryName = "UK", $stateOrProvinceName = "London", $localityName = "Blah", $organizationName = "Blah1", $organizationalUnitName = "Blah2", $commonName = "Joe Bloggs", $emailAddress = "openssl@domain.com") {
		$dn = Array ("countryName" => $countryName, "stateOrProvinceName" => $stateOrProvinceName, "localityName" => $localityName, "organizationName" => $organizationName, "organizationalUnitName" => $organizationalUnitName, "commonName" => $commonName, "emailAddress" => $emailAddress);
		$privkey = openssl_pkey_new($this->config);
		$csr = openssl_csr_new($dn, $privkey, $this->config);
		$sscert = openssl_csr_sign($csr, null, $privkey, OPEN_SSL_CERT_DAYS_VALID, $this->config);
		openssl_x509_export($sscert, $this->publickey);
		openssl_pkey_export($privkey, $this->privatekey, $this->privkeypass, $this->config);
		openssl_csr_export($csr, $this->csr);
	}

	function get_plain() {
		return $this->plaintext;
	}

	function get_crypt() {
		return $this->crypttext;
	}

	function get_ekey() {
		return $this->ekey;
	}

	function get_privatekey() {
		return $this->privatekey;
	}

	function get_privkeypass() {
		return $this->privkeypass;
	}

	function get_publickey() {
		return $this->publickey;
	}
}

?>